@@ -5,7 +5,7 @@
 #          config values from a given file and
 #          return it as hash structure
 #
-# Copyright (c) 2000-2004 Thomas Linden <tom@daemon.de>.
+# Copyright (c) 2000-2006 Thomas Linden <tom@daemon.de>.
 # All Rights Reserved. Std. disclaimer applies.
 # Artificial License, same as perl itself. Have fun.
 #
@@ -14,11 +14,20 @@ package Config::General;
 
 use FileHandle;
 use File::Spec::Functions qw(splitpath file_name_is_absolute catfile catpath);
+use File::Glob qw/:glob/;
 use strict;
+
+# on debian with perl > 5.8.4 croak() doesn't work anymore without this.
+# There is some require statement which dies 'cause it can't find Carp::Heavy,
+# I really don't understand, what the hell they made, but the debian perl
+# installation is definetly bullshit, damn!
+use Carp::Heavy;
+
+
 use Carp;
 use Exporter;
 
-$Config::General::VERSION = "2.27";
+$Config::General::VERSION = "2.31";
 
 use vars  qw(@ISA @EXPORT);
 @ISA    = qw(Exporter);
@@ -33,6 +42,8 @@ sub new {
 
   # define default options
   my $self = {
+	      SlashIsDirectory      => 0,
+
 	      AllowMultiOptions     => 1,
 
 	      MergeDuplicateOptions => 0,
@@ -42,6 +53,8 @@ sub new {
 
 	      UseApacheInclude      => 0,
 	      IncludeRelative       => 0,
+	      IncludeDirectories    => 0,
+	      IncludeGlob           => 0,
 
 	      AutoLaunder           => 0,
 
@@ -58,6 +71,8 @@ sub new {
 
 	      InterPolateVars       => 0,
 
+	      InterPolateEnv        => 0,
+
 	      ExtendedAccess        => 0,
 
 	      SplitPolicy           => 'guess', # also possible: whitespace, equalsign and custom
@@ -78,8 +93,10 @@ sub new {
 
 	      parsed                => 0,       # internal state stuff for variable interpolation
 	      upperkey              => "",
+	      upperkeys             => [],
 	      lastkey               => "",
 	      prevkey               => " ",
+	      files                 => {},      # which files we have read, if any
 	     };
 
   # create the class instance
@@ -97,13 +114,16 @@ sub new {
     $self->{ConfigFile} = delete $conf{-file} if(exists $conf{-file});
     $self->{ConfigHash} = delete $conf{-hash} if(exists $conf{-hash});
 
-    # store search path for relative configs, if any
-    $self->{ConfigPath} = delete $conf{-ConfigPath} if(exists $conf{-ConfigPath});
-
     # store input, file, handle, or array
     $self->{ConfigFile} = delete $conf{-ConfigFile} if(exists $conf{-ConfigFile});
     $self->{ConfigHash} = delete $conf{-ConfigHash} if(exists $conf{-ConfigHash});
 
+    # store search path for relative configs, if any
+    if (exists $conf{-ConfigPath}) {
+      my $configpath = delete $conf{-ConfigPath};
+      $self->{ConfigPath} = ref $configpath eq 'ARRAY' ? $configpath : [$configpath];
+    }
+
     # handle options which contains values we are needing (strings, hashrefs or the like)
     if (exists $conf{-String} ) {
       if ($conf{-String}) {
@@ -202,8 +222,10 @@ sub new {
     $self->{StoreDelimiter} = "   " if(!$self->{StoreDelimiter});
   }
 
-  if ($self->{InterPolateVars}) {
-    #
+  if ($self->{InterPolateVars} || $self->{InterPolateEnv}) {
+    # InterPolateEnv implies InterPolateVars
+    $self->{InterPolateVars} = 1;
+
     # we are blessing here again, to get into the ::InterPolated namespace
     # for inheriting the methods available overthere, which we doesn't have.
     #
@@ -298,21 +320,53 @@ sub getall {
 }
 
 
+sub files {
+  #
+  # return a list of files opened so far
+  #
+  my($this) = @_;
+  return (exists $this->{files} ? keys %{$this->{files}} : () );
+}
+
 
 sub _open {
   #
-  # open the config file
+  # open the config file, or expand a directory or glob
   #
   my($this, $configfile) = @_;
   my $fh = new FileHandle;
 
-  if( ! -e $configfile && defined($this->{ConfigPath}) ) {
-    # try to find the file within ConfigPath
-    foreach my $dir (@{$this->{ConfigPath}}) {
-      if( -e catfile($dir, $configfile) ) {
-	$configfile = catfile($dir, $configfile);
-	last; # found it
-      };
+  if ($this->{IncludeGlob} and $configfile =~ /[*?\[\{\\]/) {
+    # Something like: *.conf (or maybe dir/*.conf) was included; expand it and
+    # pass each expansion through this method again.
+    my @include = grep -f, bsd_glob($configfile, GLOB_BRACE | GLOB_QUOTE);
+    if (@include == 1) {
+      $configfile = $include[0];
+    }
+    else { # Multiple results or no expansion results (which is fine, include foo/* shouldn't fail if there isn't anything matching)
+      local $this->{IncludeGlob};
+      for (@include) {
+	$this->_open($_);
+      }
+      return;
+    }
+  }
+
+  if (!-e $configfile) {
+    my $found;
+    if (defined($this->{ConfigPath})) {
+      # try to find the file within ConfigPath
+      foreach my $dir (@{$this->{ConfigPath}}) {
+	if( -e catfile($dir, $configfile) ) {
+	  $configfile = catfile($dir, $configfile);
+	  $found = 1;
+	  last; # found it
+	}
+      }
+    }
+    if (!$found) {
+      my $path_message = defined $this->{ConfigPath} ? ' within ConfigPath: ' . join('.', @{$this->{ConfigPath}}) : '';
+      croak qq{The file "$configfile" does not exist$path_message!};
     }
   }
 
@@ -322,16 +376,36 @@ sub _open {
     $/ = "\n";
   }
 
-  if (-e $configfile) {
-    open $fh, "<$configfile" or croak "Could not open $configfile!($!)\n";
-    $this->_read($fh);
+  if (-d $configfile and $this->{IncludeDirectories}) {
+    # A directory was included; include all the files inside that directory in ASCII order
+    local *INCLUDEDIR;
+    opendir INCLUDEDIR, $configfile or croak "Could not open directory $configfile!($!)\n";
+    my @files = sort grep -f "$configfile/$_", readdir INCLUDEDIR;
+    closedir INCLUDEDIR;
+    local $this->{CurrentConfigFilePath} = $configfile;
+    for (@files) {
+      unless ($this->{files}->{"$configfile/$_"}) {
+	open $fh, "<$configfile/$_" or croak "Could not open $configfile/$_!($!)\n";
+	$this->{files}->{"$configfile/$_"} = 1;
+	$this->_read($fh);
+      }
+    }
   }
-  else {
-    if (defined $this->{ConfigPath}) {
-      croak "The file \"$configfile\" does not exist within ConfigPath: " . join(":", @{$this->{ConfigPath}}) . "!\n";
+  elsif (-e _) {
+    if (exists $this->{files}->{$configfile} ) {
+      # do not read the same file twice, just return
+      # FIXME: should we croak here, when some "debugging" is enabled?
+      return;
     }
     else {
-      croak "The file \"$configfile\" does not exist!\n";
+      open $fh, "<$configfile" or croak "Could not open $configfile!($!)\n";
+
+      $this->{files}->{$configfile}    = 1;
+
+      my ($volume, $path, undef)           = splitpath($configfile);
+      local $this->{CurrentConfigFilePath} = catpath($volume, $path, '');
+
+      $this->_read($fh);
     }
   }
 }
@@ -453,6 +527,18 @@ sub _read {
     # transform explicit-empty blocks to conforming blocks
     if (/\s*<([^\/]+?.*?)\/>$/) {
       my $block = $1;
+      if ($block !~ /\"/) {
+	if ($block !~ /\s[^\s]/) {
+	  # fix of bug 7957, add quotation to pure slash at the
+	  # end of a block so that it will be considered as directory
+	  # unless the block is already quoted or contains whitespaces
+	  # and no quotes.
+	  if ($this->{SlashIsDirectory}) {
+	    push @{$this->{content}}, '<' . $block . '"/">';
+	    next;
+	  }
+	}
+      }
       my $orig  = $_;
       $orig     =~ s/\/>$/>/;
       $block    =~ s/\s\s*.*$//;
@@ -496,7 +582,10 @@ sub _read {
       # look for include statement(s)
       my $incl_file;
       my $path = "";
-      if (defined($this->{ConfigPath})) {
+      if ( $this->{IncludeRelative} and defined($this->{CurrentConfigFilePath})) {
+      	$path = $this->{CurrentConfigFilePath};
+      }
+      elsif (defined($this->{ConfigPath})) {
 	# fetch pathname of base config file, assuming the 1st one is the path of it
 	$path = $this->{ConfigPath}->[0];
       }
@@ -530,7 +619,7 @@ sub _parse {
   # parse the contents of the file
   #
   my($this, $config, $content) = @_;
-  my(@newcontent, $block, $blockname, $grab, $chunk,$block_level);
+  my(@newcontent, $block, $blockname, $chunk,$block_level);
   local $_;
   my $indichar = chr(182);  # , inserted by _open, our here-doc indicator
 
@@ -567,9 +656,16 @@ sub _parse {
     if (! defined $block) {                                # not inside a block @ the moment
       if (/^<([^\/]+?.*?)>$/) {                            # look if it is a block
 	$block = $1;                                       # store block name
-	($grab, $blockname) = split /\s\s*/, $block, 2;    # is it a named block? if yes, store the name separately
-	if ($blockname) {
-	  $block = $grab;
+	if ($block =~ /^"([^"]+)"$/) {
+	  # quoted block, unquote it and do not split
+	  $block =~ s/"//g;
+	}
+	else {
+	  # If it is a named block store the name separately; allow the block and name to each be quoted
+	  if ($block =~ /^(?:"([^"]+)"|(\S+))(?:\s+(?:"([^"]+)"|(.*)))?$/) {
+	    $block = $1 || $2;
+	    $blockname = $3 || $4;
+	  }
 	}
 	if ($this->{InterPolateVars}) {
 	  # interpolate block(name), add "<" and ">" to the key, because
@@ -731,7 +827,7 @@ sub _parse {
 
 sub _savelast {
   my($this, $key) = @_;
-  $this->{upperkey} = $this->{lastkey};
+  push(@{$this->{upperkeys}}, $this->{lastkey});
   $this->{lastkey}  = $this->{prevkey};
   $this->{prevkey}  = $key;
 }
@@ -739,7 +835,7 @@ sub _savelast {
 sub _backlast {
   my($this, $key) = @_;
   $this->{prevkey} = $this->{lastkey};
-  $this->{lastkey} = $this->{upperkey};
+  $this->{lastkey} = pop(@{$this->{upperkeys}});
 }
 
 sub _parse_value {
@@ -965,6 +1061,11 @@ sub _write_hash {
   my $indent = "    " x $level;
   my $config_string;
 
+  if ($entry =~ /\s/) {
+    # quote the entry if it contains whitespaces
+    $entry = '"' . $entry . '"';
+  }
+
   $config_string .= $indent . "<" . $entry . ">\n";
   $config_string .= $this->_store($level + 1, %{$line});
   $config_string .= $indent . "</" . $entry . ">\n";
@@ -1188,6 +1289,22 @@ will search within B<-ConfigPath> for the file. See the description of B<-Config
 for more details.
 
 
+=item B<-IncludeDirectories>
+
+If set to a true value, you may specify include a directory, in which case all
+files inside the directory will be loaded in ASCII order.  Directory includes
+will not recurse into subdirectories.  This is comparable to including a
+directory in Apache-style config files.
+
+
+=item B<-IncludeGlob>
+
+If set to a true value, you may specify a glob pattern for an include to
+include all matching files (e.g. <<include conf.d/*.conf>>).  Also note that as
+with standard file patterns, * will not match dot-files, so <<include dir/*>>
+is often more desirable than including a directory with B<-IncludeDirectories>.
+
+
 =item B<-ConfigPath>
 
 As mentioned above, you can use this variable to specify a search path for relative
@@ -1195,7 +1312,8 @@ config files which have to be included. Config::General will search within this
 path for the file if it cannot find the file at the location relative to the
 current config file.
 
-You must specify the path as an array ref. For example:
+To provide multiple search paths you can specify an array reference for the
+path.  For example:
 
  @path = qw(/usr/lib/perl /nfs/apps/lib /home/lib);
  ..
@@ -1333,7 +1451,7 @@ which allows you to set default values for particular config options directly.
 =item B<-Tie>
 
 B<-Tie> takes the name of a Tie class as argument that each new hash should be
-based off of. 
+based off of.
 
 This hash will be used as the 'backing hash' instead of a standard perl hash,
 which allows you to affect the way, variable storing will be done. You could, for
@@ -1367,6 +1485,13 @@ Example:
 If set to a true value, variable interpolation will be done on your config
 input. See L<Config::General::Interpolated> for more informations.
 
+=item B<-InterPolateEnv>
+
+If set to a true value, environment variables can be used in
+configs.
+
+This implies B<-InterPolateVars>.
+
 =item B<-ExtendedAccess>
 
 If set to a true value, you can use object oriented (extended) methods to
@@ -1441,6 +1566,47 @@ character within configurations.
 
 By default it is turned off.
 
+
+=item B<-SlashIsDirectory>
+
+If you turn on this parameter, a single slash as the last character
+of a named block will be considered as a directory name.
+
+By default this flag is turned off, which makes the module somewhat
+incompatible to apache configs, since such a setup will be normally
+considered as an explicit empty block, just as XML defines it.
+
+For example, if you have the following config:
+
+ <Directory />
+   Index index.awk
+ </Directory>
+
+you will get such an error message from the parser:
+
+ EndBlock "</Directory>" has no StartBlock statement (level: 1, chunk 10)!
+
+This is caused by the fact that the config chunk below will be
+internally converted to:
+
+ <Directory><Directory />
+   Index index.awk
+ </Directory>
+
+Now there is one '</Directory>' too much. The proper solution is
+to use quotation to circumvent this error:
+
+ <Directory "/">
+   Index index.awk
+ </Directory>
+
+However, a raw apache config comes without such quotes. In this
+case you may consider to turn on B<-SlashIsDirectory>.
+
+Please note that this is a new option (incorporated in version 2.30),
+it may lead to various unexpected sideeffects or other failures.
+You've been warned.
+
 =back
 
 
@@ -1450,6 +1616,9 @@ By default it is turned off.
 
 Returns a hash structure which represents the whole config.
 
+=item files()
+
+Returns a list of all files read in.
 
 =item save_file()
 
@@ -1657,6 +1826,42 @@ You cannot have more than one named block with the same name because it will
 be stored in a hashref and therefore be overwritten if a block occurs once more.
 
 
+=head1 WHITESPACES IN BLOCKS
+
+The normal behavior of Config::General is to look for whitespaces in
+block names to decide if it's a named block or just a simple block.
+
+Sometimes you may need blocknames which have whitespaces in their names.
+
+With named blocks this is no problem, as the module only looks for the
+first whitespace:
+
+ <person hugo gera>
+ </person>
+
+would be parsed to:
+
+ $VAR1 = {
+          'person' => {
+                       'hugo gera' => {
+                                      },
+                      }
+         };
+
+The problem occurs, if you want to have a simple block containing whitespaces:
+
+ <hugo gera>
+ </hugo gera>
+
+This would be parsed as a named block, which is not what you wanted. In this
+very case you may use quotation marks to indicate that it is not a named block:
+
+ <"hugo gera">
+ </"hugo gera">
+
+The save() method of the module inserts automatically quotation marks in such
+cases.
+
 
 =head1 EXPICIT EMPTY BLOCKS
 
@@ -1688,7 +1893,7 @@ Example:
  log  log2
 
 You will get a scalar if the option occured only once or an array if it occured
-more than once. If you expect multiple identical options, then you may need to 
+more than once. If you expect multiple identical options, then you may need to
 check if an option occured more than once:
 
  $allowed = $hash{jonas}->{tablestructure}->{allowed};
@@ -1968,7 +2173,7 @@ I recommend you to read the following documentations, which are supplied with pe
 
 =head1 COPYRIGHT
 
-Copyright (c) 2000-2004 Thomas Linden
+Copyright (c) 2000-2006 Thomas Linden
 
 This library is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
@@ -1985,7 +2190,7 @@ Thomas Linden <tom@daemon.de>
 
 =head1 VERSION
 
-2.27
+2.31
 
 =cut
 
